import strongr.core.gateways as gateways
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship
from authlib.flask.oauth2.sqla import OAuth2TokenMixin

from strongr.core.sqlalchemydatatypes.uuidtype import UUIDType

Base = gateways.Gateways.sqlalchemy_base()

class Token(Base, OAuth2TokenMixin):
    __tablename__ = 'oauth_token'

    id = Column(Integer, primary_key=True)
    user_id = Column(String(64), ForeignKey('oauth_users.user_id', ondelete='CASCADE'))



import time

class LocalCache:
    _cache = {}
    _timeout = {}

    def _checkKey(self, key):
        if key in self._timeout and int(time.time()) > self._timeout[key]:
            del self._cache[key]
            del self._timeout[key]


    def set(self, key, value, timeout):
        self._cache[key] = value
        self._timeout[key] = int(time.time()) + timeout

    def get(self, key):
        self._checkKey(key)
        if key in self._cache:
            return self._cache[key]
        return None

    def delete(self, key):
        if self.exists(key):
            del self._timeout[key]
            del self._cache[key]

    def exists(self, key):
        self._checkKey(key)
        if key in self._cache:
            return True
        return False




import strongr
import strongr.core.domain.clouddomain
from strongr.core.gateways import Gateways
from strongr.schedulerdomain.model import Job, JobState


class StartJobOnVmHandler:
    def __call__(self, command):
        db = strongr.core.gateways.Gateways.sqlalchemy_session()
        db.query(Job).filter(Job.job_id == command.job_id).update({Job.vm_id: command.vm_id, Job.state: JobState.RUNNING}, synchronize_session='fetch')
        db.commit()

        job = db.query(Job).filter(Job.job_id == command.job_id).all()[0]

        cloudCommandBus = strongr.core.domain.clouddomain.CloudDomain.cloudService().getCommandBus()
        cloudCommandFactory = strongr.core.domain.clouddomain.CloudDomain.commandFactory()

        cloudCommandBus.handle(cloudCommandFactory.newRunJob(host=command.vm_id, image=job.image, script=job.script.splitlines(), job_id=job.job_id, scratch=job.scratch, cores=job.cores, memory=job.ram))



from strongr.schedulerdomain.model import VmState
from strongr.schedulerdomain.query import RequestScheduledJobs, RequestFinishedJobs, RequestJobInfo, FindNodeWithAvailableResources, RequestResourcesRequired, RequestVmsByState

from strongr.core.exception import InvalidParameterException

import re

class QueryFactory:
    """ This factory instantiates query objects to be sent to a scheduler querybus. """

    def newRequestVms(self, states):
        """ Generates a new RequestVms query
        :returns: A RequestVms query object
        :rtype: RequestVmsByState
        """
        if not isinstance(states, list):
            raise InvalidParameterException("states invalid")

        for state in states:
            if state not in VmState:
                raise InvalidParameterException("{} is not a valid VmState".format(state))

        return RequestVmsByState(states)

    def newRequestResourcesRequired(self):
        """ Generates a new RequestResourcesRequired query
        :returns: A RequestResourcesRequired query object
        :rtype: RequestResourcesRequired
        """
        return RequestResourcesRequired()

    def newFindNodeWithAvailableResources(self, cores, ram):
        """ Generates a new FindNodeWithAvailableResources query

        :param cores: the amount of cores needed to complete the task
        :type cores: int
        :param ram: the amount of ram needed to complete the task in GiB
        :type ram: int

        :returns: A FindNodeWithAvailableResources query object
        :rtype: FindNodeWithAvailableResources
        """
        return FindNodeWithAvailableResources(cores=cores, ram=ram)

    def newRequestScheduledJobs(self):
        """ Generates a new RequestScheduledJobs query

        :returns: A RequestScheduledJobs query object
        :rtype: RequestScheduledJobs
        """
        return RequestScheduledJobs()

    def newRequestFinishedJobs(self, jobs=None):
        """ Generates a new RequestFinishedJobs query

        :param jobs: a list of job id's

        :returns: A RequestFinishedJobs query object
        :rtype: RequestFinishedJobs
        """
        jobs_sanitized = [re.sub('[^a-zA-Z0-9-]', '', x) for x in jobs] # sanitize inputs


        return RequestFinishedJobs(jobs_sanitized)

    def newRequestJobInfo(self, jobs=None):
        """ Generates a new RequestFinishedJobs query

        :param jobs: a list of job id's

        :returns: A RequestFinishedJobs query object
        :rtype: RequestFinishedJobs
        """
        jobs_sanitized = [re.sub('[^a-zA-Z0-9-]', '', x) for x in jobs] # sanitize inputs


        return RequestJobInfo(jobs_sanitized)

    def newRequestTaskInfo(self, taskid):
        """ Generates a new RequestTaskInfo query

        :param taskid: the taskid
        :type taskid: string

        :returns: A RequestTaskInfo query object
        :rtype: RequestJobInfo
        """
        return RequestJobInfo(taskid)



import strongr.core.gateways
from strongr.schedulerdomain.model import Vm


class RequestVmsByStateHandler(object):
    def __call__(self, query):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        return session.query(Vm).filter(Vm.state.in_(query.states)).all()



from .callablecommandhandler import CallableCommandHandler
from .abstractdeployvmhandler import AbstractDeployVmHandler
from .abstractdeployvmshandler import AbstractDeployVmsHandler
from .abstractlistdeployedvmshandler import AbstractListDeployedVmsHandler
from .abstractrunjobhandler import AbstractRunJobHandler
from .abstractrequestjidstatushandler import AbstractRequestJidStatusHandler
from .abstractdestroyvmshandler import AbstractDestroyVmsHandler
from .abstractjobfinishedhandler import AbstractJobFinishedHandler



import strongr.core.domain.schedulerdomain

class CheckScalingHandler(object):
    def __call__(self, command):
        query_bus = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getQueryBus()
        command_bus = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getCommandBus()

        query_factory = strongr.core.domain.schedulerdomain.SchedulerDomain.queryFactory()
        command_factory = strongr.core.domain.schedulerdomain.SchedulerDomain.commandFactory()

        resources_required = query_bus.handle(query_factory.newRequestResourcesRequired())

        if resources_required is None or (resources_required['cores'] == 0 and resources_required['ram'] == 0):
            # scalein
            command_bus.handle(command_factory.newScaleIn())
        elif resources_required['cores'] > 0:
            # scaleout
            command_bus.handle(command_factory.newScaleOut(resources_required['cores'], resources_required['ram']))



from authlib.specs.rfc7009 import RevocationEndpoint as _RevocationEndpoint

from strongr.core.gateways import Gateways
from strongr.restdomain.model.oauth2 import Token


class RevocationEndpoint(_RevocationEndpoint):
    def query_token(self, token, token_type_hint, client):
        q = Token.query.filter_by(client_id=client.client_id)
        if token_type_hint == 'access_token':
            return q.filter_by(access_token=token).first()
        elif token_type_hint == 'refresh_token':
            return q.filter_by(refresh_token=token).first()
        # without token_type_hint
        item = q.filter_by(access_token=token).first()
        if item:
            return item
        return q.filter_by(refresh_token=token).first()

    def invalidate_token(self, token):
        session = Gateways.sqlalchemy_session()
        session.delete(token)
        session.commit()



from strongr.core.abstracts.abstractservice import AbstractService

from strongr.configdomain.command import LoadConfig
from strongr.configdomain.handler import LoadConfigHandler

class ConfigService(AbstractService):
    _command_bus = None
    _query_bus = None

    def register_models(self):
        pass

    def getCommandBus(self):
        if self._command_bus is None:
            self._command_bus = self._make_default_commandbus({
                    LoadConfigHandler: LoadConfig,
                    }, None, False)
        return self._command_bus

    def getQueryBus(self):
        if self._query_bus is None:
            self._query_bus = self._make_default_querybus({
                }, None, False)
        return self._query_bus



from cmndr import Middleware
import jsonpickle

import strongr.core

class CeleryMiddleware(Middleware):
    def __init__(self, wait_for_return):
        self._core = strongr.core.Core
        self._wait_for_return = wait_for_return

    def execute(self, command, next_callable):
        name = command.__module__.split('.')[1] + '.' + command.__class__.__name__.lower()
        command_router = self._core.command_router()
        if not command_router.has_remotable_command_registered(name):
            # if no remotable route is enabled for this command we should
            # proceed as normal
            ret = next_callable(command)
            return ret

        # a remotable route is registered, lets use it!
        result = command_router.get_remotable_command_for(name).delay(jsonpickle.encode(command))
        if self._wait_for_return:
            return result.get(timeout=500)




from flask_restplus import Namespace, Resource, fields
from flask import request, jsonify

import time
import uuid

from strongr.core.domain.schedulerdomain import SchedulerDomain
from strongr.restdomain.api.utils import namespace_require_oauth


ns = Namespace('scheduler', description='Operations related to the schedulerdomain')

post_task = ns.model('post-task', {
    'image': fields.String(required=True, min_length=1, description='The docker image where the script is executed'),
    'script': fields.String(required=True, min_length=1, description='The shellcode to be executed'),
    'scratch': fields.String(required=True, min_length=1, description='Does the job need a scratch dir?'),
    'cores': fields.Integer(required=True, min=1, description='The amount of cores needed to peform the task'),
    'memory': fields.Integer(required=True, min=1, description="The amount of ram in GiB needed to peform the task"),
    'secrets': fields.List(fields.String, required=False, description='A list of keys of secrets to be injected into the process')
})

@ns.route('/tasks/status/<string:tasks>')
class TaskStatusQuery(Resource):
    def __init__(self, *args, **kwargs):
        super(TaskStatusQuery, self).__init__(*args, **kwargs)

    @ns.response(200, 'OK')
    #@namespace_require_oauth('task')
    def get(self, tasks):
        tasks = [x.strip() for x in tasks.split(',') if len(x.strip()) > 0] # convert to array

        if tasks is None or len(tasks) == 0:
            return None, 400

        schedulerService = SchedulerDomain.schedulerService()
        queryFactory = SchedulerDomain.queryFactory()

        query = queryFactory.newRequestJobInfo(tasks)
        result = schedulerService.getQueryBus().handle(query)

        output = {}
        for job in result:
            output[job.job_id] = {
                'state': str(job.state).split('.')[-1],
                'stdout': job.stdout
            }

        return output, 200


#@ns.route('/task/<string:task_id>')
#class GetTask(Resource):
#    def __init__(self, *args, **kwargs):
#        super(GetTask, self).__init__(*args, **kwargs)
#
#    @ns.response(200, 'OK')
#    #@namespace_require_oauth('task')
#    def get(self, task_id):
#        """Requests task status"""
#        schedulerService = SchedulerDomain.schedulerService()
#        queryFactory = SchedulerDomain.queryFactory()
#
#        query = queryFactory.newRequestScheduledJobs([task_id])
#
#        result = schedulerService.getQueryBus().handle(query)
#        return result, 200

@ns.route('/task')
class Tasks(Resource):
    def __init__(self, *args, **kwargs):
        super(Tasks, self).__init__(*args, **kwargs)

    @ns.response(201, 'Task successfully created.')
    #@namespace_require_oauth('task')
    @ns.expect(post_task, validate=True)
    def post(self):
        """Creates a new task."""
        schedulerService = SchedulerDomain.schedulerService()
        commandFactory = SchedulerDomain.commandFactory()

        image = request.json['image']
        script = request.json['script'].splitlines()
        scratch = request.json['scratch']
        cores = int(request.json['cores'])
        memory = int(request.json['memory'])
        job_id = str(int(time.time())) + '-' + str(uuid.uuid4())

        if 'secrets' in request.json:
            secrets = request.json['secrets']
        else:
            secrets = []

        command = commandFactory.newScheduleJobCommand(image, script, job_id, scratch, cores, memory, secrets)

        schedulerService.getCommandBus().handle(command)
        return {'job_id': job_id}, 201



from datetime import datetime, timedelta

from sqlalchemy import and_

import strongr.core
from strongr.schedulerdomain.model import Job, JobState


class CleanupOldJobsHandler(object):
    def __call__(self, command):
        """
        MySQL:
        DELETE FROM jobs WHERE state IN ('FINISHED', 'FAILED') AND state_date < DATE_SUB(NOW(), INTERVAL 1 HOUR)
        """

        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        deadline = datetime.utcnow() - timedelta(minutes=120)

        session.delete(Job).filter(and_(Job.state.in_([JobState.FINISHED, JobState.FAILED]), Job.state_date < deadline))



from .schedulejobhandler import ScheduleJobHandler
from .runenqueuedjobshandler import RunEnqueuedJobsHandler
from .requestscheduledtaskshandler import RequestScheduledTasksHandler
from .requesttaskinfohandler import RequestTaskInfoHandler
from .findnodewithavailableresourceshandler import FindNodeWithAvailableResourcesHandler
from .startjobonvmhandler import StartJobOnVmHandler
from .checkjobsrunninghandler import CheckJobsRunningHandler
from .ensureminamountofnodeshandler import EnsureMinAmountOfNodesHandler
from .scaleouthandler import ScaleOutHandler
from .requestfinishedjobshandler import RequestFinishedJobsHandler
from .jobfinishedhandler import JobFinishedHandler
from .vmcreatedhandler import VmCreatedHandler
from .vmdestroyedhandler import VmDestroyedHandler
from .vmreadyhandler import VmReadyHandler
from .vmnewhandler import VmNewHandler
from .checkscalinghandler import CheckScalingHandler
from .requestresourcesrequiredhandler import RequestResourcesRequiredHandler
from .cleanupnodeshandler import CleanupNodesHandler
from .requestvmsbystatehandler import RequestVmsByStateHandler
from .scaleinhandler import ScaleInHandler
from .logstatshandler import LogStatsHandler
from .cleanupoldjobshandler import CleanupOldJobsHandler



from authlib.flask.oauth2.sqla import OAuth2AuthorizationCodeMixin

from strongr.core import gateways

from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from strongr.core.sqlalchemydatatypes.uuidtype import UUIDType

Base = gateways.Gateways.sqlalchemy_base()

class AuthorizationCode(Base, OAuth2AuthorizationCodeMixin):
    __tablename__ = 'oauth_auth_code'

    authorizationcode_id = Column(Integer, primary_key=True)
    user_id = Column(String(64), ForeignKey('oauth_users.user_id', ondelete='CASCADE'))



from strongr.secretsdomain.model import Secret
import strongr.core.gateways

class RemoveSecretHandler():
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        session.query(Secret).filter(Secret.key == command.key).delete()
